from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import sys

import click

from future import standard_library

from .config import read_global_config

standard_library.install_aliases()

__version__ = '2.2.0'


CONTEXT_SETTINGS = dict(auto_envvar_prefix='BM2G', help_option_names=['-h', '--help'])


class Context(object):

    def __init__(self):
        self.verbose = False
        self.home = os.getcwd()
        self._global_config = None

    @property
    def global_config(self):
        if self._global_config is None:
            self._global_config = read_global_config()
        return self._global_config

    def log(self, msg, *args):
        """Logs a message to stderr."""
        if args:
            msg %= args
        click.echo(msg, file=sys.stderr)

    def vlog(self, msg, *args):
        """Logs a message to stderr only if verbose is enabled."""
        if self.verbose:
            self.log(msg, *args)

    def exit(self, exit_code):
        self.vlog("Exiting biomaj2galaxy with exit code [%d]" % exit_code)
        sys.exit(exit_code)


pass_context = click.make_pass_decorator(Context, ensure=True)



import logging
import os
import time
import unittest

from biomaj2galaxy.cli import biomaj2galaxy

from click.testing import CliRunner

from . import gi


class DmTest(unittest.TestCase):

    def test_add_dbkey(self):

        new_dbkey = 'test_dbkey'
        new_dbkey_name = 'My cool dbkey'

        runner = CliRunner()
        runner.invoke(biomaj2galaxy, ['add', '--dbkey', new_dbkey, '--dbkey-display-name', new_dbkey_name, '--no-file-check'], catch_exceptions=False)

        dbkeys = self.gi.tool_data.show_data_table('__dbkeys__')
        dbkeys = dbkeys['fields']
        assert [new_dbkey, new_dbkey_name, ''] in dbkeys

    def test_add_bowtie2(self):

        new_dbkey = 'test_dbkey'
        new_dbkey_name = 'My cool dbkey'

        runner = CliRunner()
        runner.invoke(biomaj2galaxy, ['add', '--dbkey', new_dbkey, '--dbkey-display-name', new_dbkey_name, '--no-file-check', 'bowtie2:/some/path/foo/bar'], catch_exceptions=False)

        bowtie2 = self.gi.tool_data.show_data_table('bowtie2_indexes')
        bowtie2 = bowtie2['fields']
        assert [new_dbkey, new_dbkey, new_dbkey_name, '/some/path/foo/bar'] in bowtie2

        dbkeys = self.gi.tool_data.show_data_table('__dbkeys__')
        dbkeys = dbkeys['fields']
        assert [new_dbkey, new_dbkey_name, ''] in dbkeys

    def test_add_genome(self):

        new_dbkey = 'test_dbkey'
        new_dbkey_name = 'My cool dbkey'

        runner = CliRunner()
        runner.invoke(biomaj2galaxy, ['add', '--dbkey', new_dbkey, '--dbkey-display-name', new_dbkey_name, '--no-file-check', '-g', '/data/sample.fasta'], catch_exceptions=False)

        fasta = self.gi.tool_data.show_data_table('all_fasta')
        fasta = fasta['fields']
        assert [new_dbkey, new_dbkey, new_dbkey_name, '/galaxy-central/tool-data/test_dbkey/seq/test_dbkey.fa'] in fasta

        dbkeys = self.gi.tool_data.show_data_table('__dbkeys__')
        dbkeys = dbkeys['fields']
        assert [new_dbkey, new_dbkey_name, '/galaxy-central/tool-data/test_dbkey/len/test_dbkey.len'] in dbkeys

    def test_add_bowtie2_multiple(self):

        new_dbkey = 'test_dbkey'
        new_dbkey_name = 'My cool dbkey'

        runner = CliRunner()
        runner.invoke(biomaj2galaxy, ['add', '--dbkey', new_dbkey, '--dbkey-display-name', new_dbkey_name, '--no-file-check', 'bowtie2:/some/path/foo/bar', 'bowtie2:/some/other_path/foo/bar', 'bowtie2:/some/really/other/path/foo/bar:With a cool name'], catch_exceptions=False)

        dbkeys = self.gi.tool_data.show_data_table('__dbkeys__')
        dbkeys = dbkeys['fields']
        assert [new_dbkey, new_dbkey_name, ''] in dbkeys

        bowtie2 = self.gi.tool_data.show_data_table('bowtie2_indexes')
        bowtie2 = bowtie2['fields']
        uuids = [x[0] for x in bowtie2]
        bowtie2 = [x[1:] for x in bowtie2]
        assert [new_dbkey, new_dbkey_name, '/some/path/foo/bar'] in bowtie2
        assert [new_dbkey, new_dbkey_name, '/some/other_path/foo/bar'] in bowtie2
        assert [new_dbkey, "With a cool name", '/some/really/other/path/foo/bar'] in bowtie2

        for u in uuids:
            assert u.startswith(new_dbkey)

    def test_add_mess(self):

        new_dbkey = 'test_dbkey'
        new_dbkey_name = 'My cool dbkey'

        runner = CliRunner()
        runner.invoke(biomaj2galaxy, ['add', '--dbkey', new_dbkey, '--dbkey-display-name', new_dbkey_name, '--no-file-check', 'bowtie2:/some/path/foo/bar', 'bowtie2:/some/other_path/foo/bar', 'bowtie2:/some/really/other/path/foo/bar:With a cool name', 'blastdb:/foo/really/other/path/foo/bar:With a cool name too!', 'blastdb:/foo/really/other/xxxx/bar:Wisuith a cool name too!', 'bwa:/foo/really/foo/bar:With a cool name too bwa!', 'twobit:/foo/really/foo/bar:Wblablatoo!', 'star:/foo/bloup/test/faa/bor:Wixxxtoo!', 'fasta:/fxx/bloup/txx/faa/bor:Wixxx fasta too!'], catch_exceptions=False)

        dbkeys = self.gi.tool_data.show_data_table('__dbkeys__')
        dbkeys = dbkeys['fields']
        assert [new_dbkey, new_dbkey_name, ''] in dbkeys

        bowtie2 = self.gi.tool_data.show_data_table('bowtie2_indexes')
        bowtie2 = bowtie2['fields']
        uuids = [x[0] for x in bowtie2]
        bowtie2 = [x[1:] for x in bowtie2]
        assert [new_dbkey, new_dbkey_name, '/some/path/foo/bar'] in bowtie2
        assert [new_dbkey, new_dbkey_name, '/some/other_path/foo/bar'] in bowtie2
        assert [new_dbkey, "With a cool name", '/some/really/other/path/foo/bar'] in bowtie2

        for u in uuids:
            assert u.startswith(new_dbkey) and u != new_dbkey

        blastdb = self.gi.tool_data.show_data_table('blastdb')
        blastdb = blastdb['fields']
        uuids = [x[0] for x in blastdb]
        blastdb = [x[1:] for x in blastdb]
        assert ['With a cool name too!', '/foo/really/other/path/foo/bar'] in blastdb
        assert ["Wisuith a cool name too!", '/foo/really/other/xxxx/bar'] in blastdb

        for u in uuids:
            assert u.startswith(new_dbkey) and u != new_dbkey

        bwa = self.gi.tool_data.show_data_table('bwa_indexes')
        bwa = bwa['fields']
        assert [new_dbkey, new_dbkey, 'With a cool name too bwa!', '/foo/really/foo/bar'] in bwa

        twobit = self.gi.tool_data.show_data_table('twobit')
        twobit = twobit['fields']
        assert [new_dbkey, '/foo/really/foo/bar'] in twobit

        star = self.gi.tool_data.show_data_table('rnastar_index2x_versioned')
        star = star['fields']
        assert [new_dbkey, new_dbkey, 'Wixxxtoo!', '/foo/bloup/test/faa/bor', '0', '0'] in star

        fasta = self.gi.tool_data.show_data_table('all_fasta')
        fasta = fasta['fields']
        assert [new_dbkey, new_dbkey, 'Wixxx fasta too!', '/fxx/bloup/txx/faa/bor'] in fasta

    def test_add_star_gtf(self):

        new_dbkey = 'test_dbkey'
        new_dbkey_name = 'My cool dbkey'

        runner = CliRunner()
        runner.invoke(biomaj2galaxy, ['add', '--dbkey', new_dbkey, '--dbkey-display-name', new_dbkey_name, '--no-file-check', 'star:/foo/bloup/test/faa/bor:Wixxxtoo!', '--star-with-gtf'], catch_exceptions=False)

        star = self.gi.tool_data.show_data_table('rnastar_index2x_versioned')
        star = star['fields']
        assert [new_dbkey, new_dbkey, 'Wixxxtoo!', '/foo/bloup/test/faa/bor', '1', '0'] in star

    def test_add_star_gtf_version(self):

        new_dbkey = 'test_dbkey'
        new_dbkey_name = 'My cool dbkey'

        runner = CliRunner()
        runner.invoke(biomaj2galaxy, ['add', '--dbkey', new_dbkey, '--dbkey-display-name', new_dbkey_name, '--no-file-check', 'star:/foo/bloup/test/faa/bor:Wixxxtoo!', '--star-with-gtf', '--star-version', '1.7.6'], catch_exceptions=False)

        star = self.gi.tool_data.show_data_table('rnastar_index2x_versioned')
        star = star['fields']
        assert [new_dbkey, new_dbkey, 'Wixxxtoo!', '/foo/bloup/test/faa/bor', '1', '1.7.6'] in star

    def test_add_biomaj_env(self):

        new_dbkey = 'test_dbkey'
        new_dbkey_name = 'Cool_db (v3.0)'

        back_env = os.environ
        os.environ['dbname'] = "Cool_db"
        os.environ['remoterelease'] = "v3.0"
        os.environ['data.dir'] = "/db/"
        os.environ['dirversion'] = "some/dir"
        os.environ['localrelease'] = "v3.0beta"

        runner = CliRunner()
        runner.invoke(biomaj2galaxy, ['add', '--dbkey', new_dbkey, '--no-file-check', 'bowtie2:foo/bar'], catch_exceptions=False)

        dbkeys = self.gi.tool_data.show_data_table('__dbkeys__')
        dbkeys = dbkeys['fields']
        assert [new_dbkey, new_dbkey_name, ''] in dbkeys

        bowtie2 = self.gi.tool_data.show_data_table('bowtie2_indexes')
        bowtie2 = bowtie2['fields']
        assert [new_dbkey, new_dbkey, new_dbkey_name, '/db/some/dir/v3.0beta/foo/bar'] in bowtie2

        os.environ = back_env

    def test_rm_bowtie2(self):

        new_dbkey = 'test_dbkey'
        new_dbkey_name = 'My cool dbkey'

        runner = CliRunner()
        runner.invoke(biomaj2galaxy, ['add', '--dbkey', new_dbkey, '--dbkey-display-name', new_dbkey_name, '--no-file-check', 'bowtie2:/some/path/foo/bar'], catch_exceptions=False)

        dbkeys = self.gi.tool_data.show_data_table('__dbkeys__')
        dbkeys = dbkeys['fields']
        assert [new_dbkey, new_dbkey_name, ''] in dbkeys

        bowtie2 = self.gi.tool_data.show_data_table('bowtie2_indexes')
        bowtie2 = bowtie2['fields']
        assert [new_dbkey, new_dbkey, new_dbkey_name, '/some/path/foo/bar'] in bowtie2

        runner.invoke(biomaj2galaxy, ['rm', new_dbkey], catch_exceptions=False)

        dbkeys = self.gi.tool_data.show_data_table('__dbkeys__')
        dbkeys = dbkeys['fields']
        assert [new_dbkey, new_dbkey_name, ''] not in dbkeys

        bowtie2 = self.gi.tool_data.show_data_table('bowtie2_indexes')
        bowtie2 = bowtie2['fields']
        assert [new_dbkey, new_dbkey, new_dbkey_name, '/some/path/foo/bar'] not in bowtie2

    def test_rm_multiple(self):

        new_dbkey = 'test_dbkey'
        new_dbkey_name = 'My cool dbkey'

        runner = CliRunner()
        runner.invoke(biomaj2galaxy, ['add', '--dbkey', new_dbkey, '--dbkey-display-name', new_dbkey_name, '--no-file-check', 'bowtie2:/some/path/foo/bar', 'bowtie2:/some/other_path/foo/bar', 'bowtie2:/some/really/other/path/foo/bar:With a cool name'], catch_exceptions=False)

        dbkeys = self.gi.tool_data.show_data_table('__dbkeys__')
        dbkeys = dbkeys['fields']
        assert [new_dbkey, new_dbkey_name, ''] in dbkeys

        bowtie2 = self.gi.tool_data.show_data_table('bowtie2_indexes')
        bowtie2 = bowtie2['fields']
        uuids = [x[0] for x in bowtie2]
        bowtie2 = [x[1:] for x in bowtie2]
        assert [new_dbkey, new_dbkey_name, '/some/path/foo/bar'] in bowtie2
        assert [new_dbkey, new_dbkey_name, '/some/other_path/foo/bar'] in bowtie2
        assert [new_dbkey, "With a cool name", '/some/really/other/path/foo/bar'] in bowtie2

        for u in uuids:
            assert u.startswith(new_dbkey)

        runner.invoke(biomaj2galaxy, ['rm', new_dbkey], catch_exceptions=False)

        dbkeys = self.gi.tool_data.show_data_table('__dbkeys__')
        dbkeys = dbkeys['fields']
        assert [new_dbkey, new_dbkey_name, ''] not in dbkeys

        bowtie2 = self.gi.tool_data.show_data_table('bowtie2_indexes')
        bowtie2 = bowtie2['fields']
        uuids = [x[0] for x in bowtie2]
        bowtie2 = [x[1:] for x in bowtie2]
        assert [new_dbkey, new_dbkey_name, '/some/path/foo/bar'] not in bowtie2
        assert [new_dbkey, new_dbkey_name, '/some/other_path/foo/bar'] not in bowtie2
        assert [new_dbkey, "With a cool name", '/some/really/other/path/foo/bar'] not in bowtie2

        for u in uuids:
            assert not u.startswith(new_dbkey)

    def test_rm_multiple_exact(self):

        new_dbkey = 'test_dbkey'
        new_dbkey_name = 'My cool dbkey'

        runner = CliRunner()
        runner.invoke(biomaj2galaxy, ['add', '--dbkey', new_dbkey, '--dbkey-display-name', new_dbkey_name, '--no-file-check', 'bowtie2:/some/path/foo/bar', 'bowtie2:/some/other_path/foo/bar', 'bowtie2:/some/really/other/path/foo/bar:With a cool name'], catch_exceptions=False)

        dbkeys = self.gi.tool_data.show_data_table('__dbkeys__')
        dbkeys = dbkeys['fields']
        assert [new_dbkey, new_dbkey_name, ''] in dbkeys

        bowtie2 = self.gi.tool_data.show_data_table('bowtie2_indexes')
        bowtie2 = bowtie2['fields']
        uuids = [x[0] for x in bowtie2]
        bowtie2 = [x[1:] for x in bowtie2]
        assert [new_dbkey, new_dbkey_name, '/some/path/foo/bar'] in bowtie2
        assert [new_dbkey, new_dbkey_name, '/some/other_path/foo/bar'] in bowtie2
        assert [new_dbkey, "With a cool name", '/some/really/other/path/foo/bar'] in bowtie2

        for u in uuids:
            assert u.startswith(new_dbkey)

        runner.invoke(biomaj2galaxy, ['rm', '--exact', new_dbkey], catch_exceptions=False)

        dbkeys = self.gi.tool_data.show_data_table('__dbkeys__')
        dbkeys = dbkeys['fields']
        assert [new_dbkey, new_dbkey_name, ''] not in dbkeys

        bowtie2 = self.gi.tool_data.show_data_table('bowtie2_indexes')
        bowtie2 = bowtie2['fields']
        uuids = [x[0] for x in bowtie2]
        bowtie2 = [x[1:] for x in bowtie2]
        assert [new_dbkey, new_dbkey_name, '/some/path/foo/bar'] not in bowtie2
        assert [new_dbkey, new_dbkey_name, '/some/other_path/foo/bar'] not in bowtie2
        assert [new_dbkey, "With a cool name", '/some/really/other/path/foo/bar'] not in bowtie2

        for u in uuids:
            assert u.startswith(new_dbkey) and u != new_dbkey

    def setUp(self):
        self.gi = gi

        logging.getLogger("urllib3").setLevel(logging.WARNING)
        logging.getLogger("bioblend").setLevel(logging.WARNING)

        # Empty all tables
        touched_tables = [
            'bowtie2_indexes',
            'blastdb',
            'twobit',
            '__dbkeys__',
            'rnastar_index2x_versioned',
            'bwa_indexes',
            'all_fasta'
        ]
        tables = [x['name'] for x in self.gi.tool_data.get_data_tables()]
        for table in tables:
            if table in touched_tables:  # To speed things up a bit
                fields = self.gi.tool_data.show_data_table(table)['fields']
                for line in fields:
                    self.gi.tool_data.delete_data_table(table, "\t".join(line))
                time.sleep(1)  # Reloading too soon might not work for some strange reason
                self.gi.tool_data.reload_data_table(table)

    def tearDown(self):
        # Empty all tables
        touched_tables = [
            'bowtie2_indexes',
            'blastdb',
            'twobit',
            '__dbkeys__',
            'rnastar_index2x_versioned',
            'bwa_indexes',
            'all_fasta'
        ]
        tables = [x['name'] for x in self.gi.tool_data.get_data_tables()]
        for table in tables:
            if table in touched_tables:  # To speed things up a bit
                fields = self.gi.tool_data.show_data_table(table)['fields']
                for line in fields:
                    self.gi.tool_data.delete_data_table(table, "\t".join(line))
                time.sleep(1)  # Reloading too soon might not work for some strange reason
                self.gi.tool_data.reload_data_table(table)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os

from biomaj2galaxy import pass_context
from biomaj2galaxy.utils import add_files, check_existing, check_input, create_tree, get_library, get_roles

import click


@click.command()
@click.argument("sources", nargs=-1, type=click.Path())
@click.option(
    "-l",
    "--library",
    help="Name of the destination library (default=guessed from BioMAJ bank name, ie $dbname env var)",
    type=str
)
@click.option(
    "-f",
    "--folder",
    help="Data library folder where the data will be placed (default=/)",
    default="/",
    type=str
)
@click.option(
    "--roles",
    help="Restrict acces to given role(s) (comma separated list). WARNING: the permission of the whole library is modified when using this option.",
    default="",
    type=str
)
@click.option(
    "--lib-desc",
    help="Library description (only used if the library does not already exist)",
    default="",
    type=str
)
@click.option(
    "--lib-synopsis",
    help="Library synopsis (only used if the library does not already exist)",
    default="",
    type=str
)
@click.option(
    "--datatype",
    help="Datatype of the file(s) to add to the data library (default=auto detect)",
    default="auto",
    type=str
)
@click.option(
    "--no-file-check",
    help="Don't check the source files existence.\nUseful for files that are available on the Galaxy server, but not on the machine running this script.",
    is_flag=True
)
@click.option(
    "--replace",
    help="Replace files already present in the data library (with the same name)",
    is_flag=True
)
@click.option(
    "--no-biomaj-env",
    help="Add this flag if you don't want biomaj2galaxy to use BioMAJ env variables to guess file names.",
    is_flag=True
)
@pass_context
def add_lib(ctx, sources, library, folder, roles, lib_desc, lib_synopsis, datatype, no_file_check, replace, no_biomaj_env):
    """Add data to a Galaxy data library, where SOURCES a list of file/directories to add."""

    if not sources:
        print("Nothing to do")
        return

    if not library:
        if 'dbname' in os.environ:
            library = os.environ['dbname']
        else:
            raise Exception('No library defined. Use the --library option.')

    sources = check_input(sources, check_existence=(not no_file_check), use_biomaj_env=(not no_biomaj_env))

    r_roles = []
    if roles:
        print("Checking roles")
        roles = roles.split(',')
        r_roles = get_roles(ctx.gi, roles)

    print("Adding to data library '" + str(library) + "'")

    if not folder:
        folder = '/'
    dest = os.path.normpath(folder)
    dest = dest.split(os.sep)
    dest = [x for x in dest if x]  # Remove empty string when sep at the begin or end, or multiple sep

    found_lib = get_library(ctx.gi, library, lib_desc, lib_synopsis)

    print("Preparing folders in library '" + library + "'")

    dest_folder = create_tree(ctx.gi, found_lib, dest)

    check_existing(ctx.gi, found_lib, dest, sources, replace)

    print("Adding " + str(len(sources)) + " file(s) to the library '" + library + "'")
    add_files(ctx.gi, found_lib, dest_folder, sources, r_roles, datatype)

    print("Done!")



from biomaj2galaxy import __version__
from biomaj2galaxy import pass_context

import click

from .commands.add import add as func0
from .commands.add_lib import add_lib as func2
from .commands.init import init as func4
from .commands.rm import rm as func1
from .commands.rm_lib import rm_lib as func3
from .config import get_instance, global_config_path, set_global_config_path


@click.group()
@click.version_option(__version__)
@click.option('-v', '--verbose', is_flag=True,
              help='Enables verbose mode.')
@click.option(
    "-i",
    "--instance",
    help='Name of instance in %s. This parameter can also be set via the environment variable BM2G_INSTANCE' % global_config_path(),
    default='__default',
    show_default=True,
    required=True
)
@click.option(
    "--path", "-f",
    help="config file path",
    type=str
)
@pass_context
def biomaj2galaxy(ctx, instance, verbose, path=None):
    # set config_path if provided
    if path is not None and len(path) > 0:
        set_global_config_path(path)
    # We abuse this, knowing that calls to one will fail.
    current_ctx = click.get_current_context()
    ctx.gi = None
    try:
        ctx.gi = get_instance(instance)
    except TypeError:
        pass

    if current_ctx.invoked_subcommand not in ['init'] and ctx.gi is None:
        raise Exception("Could not read config file '%s', run `biomaj2galaxy init` to create it, or set BM2G_GLOBAL_CONFIG_PATH to the correct location." % global_config_path())

    ctx.verbose = verbose


biomaj2galaxy.add_command(func0)
biomaj2galaxy.add_command(func1)
biomaj2galaxy.add_command(func2)
biomaj2galaxy.add_command(func3)
biomaj2galaxy.add_command(func4)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os

from biomaj2galaxy import pass_context
from biomaj2galaxy.utils import find_tree, get_library

import click


@click.command()
@click.argument("library", nargs=1)
@click.option(
    "-f",
    "--folder",
    help="Data library folder to remove (default=remove the whole library)",
    type=str
)
@pass_context
def rm_lib(ctx, library, folder):
    """Remove data from the LIBRARY Galaxy data library"""

    print("Removing from data library '" + str(library) + "'")

    found_lib = get_library(ctx.gi, library)

    if folder:
        dest = os.path.normpath(folder)
        dest = dest.split(os.sep)
        dest = [x for x in dest if x]  # Remove empty string when sep at the begin or end, or multiple sep

        print("Looking for folder '" + folder + "' in library '" + library + "'")
        dest_folder = find_tree(ctx.gi, found_lib, dest)

        print("Removing folder '" + folder + "' from the library '" + library + "'")
        ctx.gi.folders.delete_folder(dest_folder, True)

    else:
        print("Removing the whole library '" + library + "'")
        ctx.gi.libraries.delete_library(found_lib)

    print("Done!")



from __future__ import absolute_import

import os

from bioblend import galaxy

import yaml


DEFAULT_CONFIG = {
}

_config_path = os.environ.get(
    "BM2G_GLOBAL_CONFIG_PATH",
    "~/.bm2g.yml"
)
_config_path = os.path.expanduser(_config_path)
DEFAULT_CONFIG['config_path'] = _config_path


def global_config_path():
    return DEFAULT_CONFIG['config_path']


def set_global_config_path(config_path):
    DEFAULT_CONFIG['config_path'] = config_path


def read_global_config():
    config_path = global_config_path()
    if not os.path.exists(config_path):
        return DEFAULT_CONFIG

    with open(config_path) as f:
        return yaml.load(f)


def _get_instance(instance_name=None):
    # I don't like reading the config twice.
    conf = read_global_config()

    if not os.path.exists(global_config_path()):
        # Probably creating the file for the first time.
        return None

    if instance_name is None or instance_name == '__default':
        try:
            instance_name = conf['__default']
        except KeyError:
            raise Exception("Unknown Galaxy instance and no __default provided")

    if instance_name not in conf:
        raise Exception("Unknown Galaxy instance; check spelling or add to %s" % DEFAULT_CONFIG)

    return conf[instance_name]


def get_instance(instance_name=None, offline=False):
    conf = _get_instance(instance_name=instance_name)
    return galaxy.GalaxyInstance(url=conf['url'],
                                 key=conf['apikey'],)



# coding: utf-8
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os

from bioblend import galaxy

from biomaj2galaxy import config, pass_context
from biomaj2galaxy.io import info, warn

import click


CONFIG_TEMPLATE = """## BioMAJ2Galaxy: Global Configuration File.
# Each stanza should contain a single Galaxy server to interact with.
#
# You can set the key __default to the name of a default instance
__default: local

local:
    url: "%(url)s"
    apikey: "%(apikey)s"
"""

SUCCESS_MESSAGE = (
    "Ready to go! Type `biomaj2galaxy` to get a list of commands you can execute."
)


@click.command()
@pass_context
def init(ctx, url=None, api_key=None, admin=False, **kwds):
    """Help initialize global configuration (in home directory)
    """

    click.echo("""Welcome to BioMAJ2Galaxy""")
    if os.path.exists(config.global_config_path()):
        info("Your biomaj2galaxy configuration already exists. Please edit it instead: %s" % config.global_config_path())
        return 0

    while True:
        # Check environment
        url = click.prompt("url")
        apikey = click.prompt("apikey")

        info("Testing connection...")
        try:
            instance = galaxy.GalaxyInstance(url=url, key=apikey)
            instance.libraries.get_libraries()
            # We do a connection test during startup.
            info("Ok! Everything looks good.")
            break
        except Exception as e:
            warn("Error, we could not access the configuration data for your instance: %s", e)
            should_break = click.prompt("Continue despite inability to contact this instance? [y/n]")
            if should_break in ('Y', 'y'):
                break

    config_path = config.global_config_path()
    if os.path.exists(config_path):
        warn("File %s already exists, refusing to overwrite." % config_path)
        return -1

    with open(config_path, "w") as f:
        f.write(CONFIG_TEMPLATE % {
            'url': url,
            'apikey': apikey,
        })
        info(SUCCESS_MESSAGE)



from __future__ import absolute_import

import click


def info(message, *args):
    if args:
        message = message % args
    click.echo(click.style(message, bold=True, fg='green'), err=True)


def error(message, *args):
    if args:
        message = message % args
    click.echo(click.style(message, bold=True, fg='red'), err=True)


def warn(message, *args):
    if args:
        message = message % args
    click.echo(click.style(message, fg='yellow'), err=True)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import fnmatch
import os
import sys
import time

from bioblend import ConnectionError

from biomaj2galaxy.io import warn


def get_roles(gi, roles):
    """
    Find role ids corresponding to the ones given with -r option
    """
    remotes = gi.roles.get_roles()
    remotes = {r['name']: r['id'] for r in remotes}

    ids = []
    for a in roles:
        if a not in remotes:
            raise Exception("Could not find role '" + a + "'")
        else:
            ids.append(remotes[a])

    return ids


def isfile_wo_ext(path):
    """Checks that a file exists, or exists with any extension"""
    path_dir = os.path.dirname(path)
    path_fn = os.path.basename(path)

    return os.path.isfile(path) or len(fnmatch.filter(os.listdir(path_dir), path_fn + '.*')) > 0


def check_input(sources, check_existence=True, use_biomaj_env=True):
    formatted_source = []
    print("Checking input files, converting to absolute path: %s" % sources)
    for f in sources:
        if use_biomaj_env and not f.startswith('/') and 'data.dir' in os.environ and 'dirversion' in os.environ and 'localrelease' in os.environ:
            abs_path = os.path.join(os.environ['data.dir'], os.environ['dirversion'], os.environ['localrelease'], f)
        else:
            abs_path = os.path.abspath(f)

        if check_existence and not isfile_wo_ext(abs_path):
            raise Exception("File '" + abs_path + "' could not be read!")

        formatted_source.append(abs_path)
    return formatted_source


def get_library(gi, lib_name, lib_desc="", lib_synopsis="", create=True):
    """
    Get the id corresponding to given library, and create it if it doesn't exist yet
    """
    print("Looking for lib '" + lib_name + "'")
    libs = gi.libraries.get_libraries()

    found_lib = None
    for lib in libs:
        if not found_lib and lib['name'] == lib_name and lib['deleted'] is False:
            print("Found library '" + lib_name + "'")
            found_lib = lib['id']

    if not found_lib:
        if create:
            print("Did not find library '" + lib_name + "', creating it")
            create = gi.libraries.create_library(lib_name, lib_desc, lib_synopsis)
            found_lib = create['id']
        else:
            raise Exception("Did not find library '" + lib_name + "'")

    return found_lib


def find_tree(gi, found_lib, folders):
    """
    Look for a directory structure in the given library.
    Returns the id of the last folder of the tree if it was completely found
    """
    dist_folders = gi.libraries.get_folders(found_lib)

    dist_f = {}
    for f in dist_folders:
        dist_f[f['name']] = f

    path = ""
    last_f_id = None
    for f in folders:
        path += "/" + f
        if path in dist_f:
            print("Found folder " + f)
            last_f_id = dist_f[path]['id']
        else:
            raise Exception("Did not find folder '" + f + "'")

    return last_f_id


def create_tree(gi, found_lib, folders):
    """
    Create a directory structure in the given library.
    Returns the id of the last folder of the tree
    """
    dist_folders = gi.libraries.get_folders(found_lib)

    dist_f = {}
    for f in dist_folders:
        dist_f[f['name']] = f

    path = ""
    last_f_id = None
    folder_to_create = []
    for f in folders:
        path += "/" + f
        if path in dist_f:
            print("Found folder " + f)
            last_f_id = dist_f[path]['id']
        else:
            print("Did not find folder " + f)
            folder_to_create.append(f)

    if len(folder_to_create) > 0:
        for f in folder_to_create:
            if last_f_id:
                print("Creating folder " + f + " in folder " + last_f_id)
                f_c = gi.libraries.create_folder(found_lib, f, "", last_f_id)
            else:
                print("Creating folder " + f + " in root folder")
                f_c = gi.libraries.create_folder(found_lib, f, "")
            last_f_id = f_c[0]['id']

    return last_f_id


def add_files(gi, lib, dest_folder, source, roles, file_type):
    for f in source:
        gi.libraries.upload_from_galaxy_filesystem(lib, f, dest_folder, file_type, link_data_only='link_to_files')
        if roles:
            gi.libraries.set_library_permissions(lib, access_in=roles)


def check_existing(gi, lib, dest, source, replace):

    dest = "/" + "/".join(dest)
    files = []
    for f in source:
        files.append(dest + "/" + os.path.basename(f))

    libs = gi.libraries.show_library(lib, True)
    for e in libs:
        if e['type'] == 'file' and e['name'] in files:
            if replace:
                print(e['name'] + " already present in the data library: replacing it")
                gi.libraries.delete_library_dataset(lib, e['id'])
            else:
                print(e['name'] + " already present in the data library: adding another copy")


def get_dbkey_entry(gi, dbkey):
    dbkeys = gi.tool_data.show_data_table('__dbkeys__')

    for k in dbkeys['fields']:
        if k[0] == dbkey:
            return k


def wait_completion(gi, dataset_id, job_id, sleep_time=30, exit_on_error=True):
    error_number = 0
    sleep_time = 10  # Don't wait too much for first try

    dataset_info = []
    while True:
        # What's the status of the running job?
        try:
            dataset_info = gi.datasets.show_dataset(dataset_id)
        except ConnectionError:
            error_number += 1
            warn("Could not connect to the Galaxy server, waiting %s seconds before retrying..." % (sleep_time * error_number))
            time.sleep(sleep_time * (error_number - 1))  # -1 because we already wait a little below

            if error_number > 50:
                raise Exception('Could not connect to the Galaxy server for too long, giving up.')

        # Finished!
        if dataset_info and 'state' in dataset_info and dataset_info['state'] in ['ok', 'error']:
            break

        # Not finished yet, wait a little
        time.sleep(sleep_time)
        sleep_time = 30

    dataset_info = gi.datasets.show_dataset(dataset_id)
    status = dataset_info['state']
    if exit_on_error and status == 'error' and job_id is not None:
        details = gi.jobs.show_job(job_id, full_details=True)
        print("STDOUT content:")
        print(details['stdout'])
        print("STDERR content:", file=sys.stderr)
        print(details['stderr'], file=sys.stderr)
        raise Exception("Job finished in error state! Aborting")



# http://bugs.python.org/issue15881#msg170215
from setuptools import find_packages, setup

setup(
    name="biomaj2galaxy",
    version='2.2.0',
    description="Command-line utility to assist in interconnecting BioMAJ (https://biomaj.genouest.org/) with Galaxy (http://galaxyproject.org/).",
    author="Anthony Bretaudeau",
    author_email="anthony.bretaudeau@inra.fr",
    url="https://github.com/genouest/biomaj2galaxy",
    install_requires=['future', 'bioblend', 'click'],
    packages=find_packages(),
    license='MIT',
    platforms="Posix; MacOS X; Windows",
    entry_points='''
        [console_scripts]
        biomaj2galaxy=biomaj2galaxy.cli:biomaj2galaxy
    ''',
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ])



import os
import time
import uuid

from biomaj2galaxy import pass_context
from biomaj2galaxy.io import warn
from biomaj2galaxy.utils import check_input, get_dbkey_entry, wait_completion

import click


@click.command()
@click.argument("files", nargs=-1)
@click.option(
    "-d",
    "--dbkey",
    help="Dbkey to use (i.e. genome build like \'hg19\')",
    type=str
)
@click.option(
    "-n",
    "--dbkey-display-name",
    help="Display name for the dbkey (default=guessed from BioMAJ env vars, ie '${dbname} (${remoterelease})')",
    type=str
)
@click.option(
    "-g",
    "--genome-fasta",
    help="Path to a fasta file corresponding to a full reference genome. It will be used in visualizations for example.",
    type=str
)
@click.option(
    "--genome-fasta-name",
    help="Display name for the full reference genome (default=--dbkey-display-name or --dbkey)",
    type=str
)
@click.option(
    "-s",
    "--fasta-sorting-method",
    help="Method used for the sorting of genome fasta file",
    type=click.Choice(['as_is', 'lexicographical', 'gatk', 'custom']),
    default='as_is'
)
@click.option(
    "--fasta-custom-sort-list",
    help="Ordered comma separated list of sequence identifiers to use for sorting genome fasta file (requires \'-s custom\' option)",
    type=str
)
@click.option(
    "--fasta-custom-sort-handling",
    help="How to handle non-specified identifiers (requires \'-s custom\' option)",
    type=click.Choice(['discard', 'keep_append', 'keep_prepend']),
    default='discard'
)
@click.option(
    "--no-file-check",
    help="Don't check the source files existence.\nUseful for files that are available on the Galaxy server, but not on the machine running this script.",
    is_flag=True
)
@click.option(
    "--star-with-gtf",
    help="STAR indices were made with an annotation (i.e., --sjdbGTFfile and --sjdbOverhang were used).",
    is_flag=True
)
@click.option(
    "--star-version",
    help="Version of STAR used to create the index (default: none)",
    type=str
)
@click.option(
    "--no-biomaj-env",
    help="Add this flag if you don't want biomaj2galaxy to use BioMAJ env variables to guess file names.",
    is_flag=True
)
@pass_context
def add(ctx, files, dbkey, dbkey_display_name, genome_fasta, genome_fasta_name, fasta_sorting_method, fasta_custom_sort_list, fasta_custom_sort_handling, no_file_check, star_with_gtf, star_version, no_biomaj_env):
    """Add data to a Galaxy data table. FILES is a list of path respecting this syntax: data_table_name:/path/to/data:Data name (e.g. "bowtie2:/db/some/where/my_genome:My supercool genome"). You can escape ':' by writing '\\:'"""

    ADD_FASTA_TOOL_ID = 'toolshed.g2.bx.psu.edu/repos/devteam/data_manager_fetch_genome_dbkeys_all_fasta/data_manager_fetch_genome_all_fasta_dbkey/0.0.4'
    DM_MANUAL_TOOL_ID = 'toolshed.g2.bx.psu.edu/repos/iuc/data_manager_manual/data_manager_manual/0.0.2'

    # Fetch the list of known tables with their columns
    tables_format = {}
    tables = ctx.gi.tool_data.get_data_tables()
    for t in tables:
        tables_format[t['name']] = ctx.gi.tool_data.show_data_table(t['name'])['columns']

        # A stupid fix for the twobit table which for some unknown reason doesn't have a 'name' column_name
        # As this 'name' column is required for a data table, the galaxy code adds a non-existing one when it is not found in the table defintion.
        if t['name'] == 'twobit' and 'name' in tables_format[t['name']]:
            tables_format[t['name']].remove('name')

    # Define some simpler synonyms for data tables
    data_table_synonyms = {
        'fasta': 'all_fasta',
        'bowtie': 'bowtie_indexes',
        'bowtie2': 'bowtie2_indexes',
        'bwa': 'bwa_indexes',
        'bwa_mem': 'bwa_mem_indexes',
        'tophat2': 'tophat2_indexes',
        'star': 'rnastar_index2x_versioned',
    }

    files_info = []
    for f in files:
        f = f.replace("\\:", '___colon___')
        f_info = f.split(':')
        f_info = [x.replace('___colon___', ':') for x in f_info]

        if len(f_info) < 2 or len(f_info) > 3:
            raise Exception('Malformed file information "%s"' % f_info)

        if f_info[0] not in tables_format:
            if f_info[0] in data_table_synonyms:
                f_info[0] = data_table_synonyms[f_info[0]]
            else:
                raise Exception('Unknown data table name "%s"' % f_info[0])

        f_info[1] = check_input([f_info[1]], check_existence=(not no_file_check), use_biomaj_env=(not no_biomaj_env))[0]

        if len(f_info) == 3:
            files_info.append({'table': f_info[0], 'path': f_info[1], 'name': f_info[2]})
        else:
            files_info.append({'table': f_info[0], 'path': f_info[1]})

    # Check which tables we're touching
    table_counts = {}
    for f_info in files_info:
        if f_info['table'] not in table_counts:
            table_counts[f_info['table']] = 0

        table_counts[f_info['table']] += 1

    # Verify dbkey
    dbkey_entry = get_dbkey_entry(ctx.gi, dbkey)
    dbkey_exists = dbkey_entry is not None

    need_dbkey = genome_fasta or (len(table_counts) == 0)
    for c in table_counts.keys():
        need_dbkey = 'dbkey' in tables_format[c]
        if need_dbkey:
            break

    create_dbkey = dbkey and not dbkey_exists and need_dbkey

    if create_dbkey:
        print("Need to create the dbkey '" + dbkey + "'")
    elif dbkey and dbkey_exists:
        print("The dbkey '" + dbkey + "' already exists")
    elif not dbkey and not need_dbkey:
        print("No dbkey was specified, but it is not a problem as we don't need it.")
    elif not dbkey and need_dbkey:
        raise Exception("ERROR: You must specify a dbkey to perform the action(s) you requested.")

    # Prepare a default display name that will be used if not specified
    if not dbkey_display_name:
        if 'dbname' in os.environ and 'remoterelease' in os.environ:
            dbkey_display_name = "%s (%s)" % (os.environ['dbname'], os.environ['remoterelease'])

    default_name = dbkey_display_name
    if not default_name and dbkey_entry:
        print("Trying to use dbkey_entry name: " + dbkey_entry[1])
        default_name = dbkey_entry[1]
    if not default_name:
        default_name = dbkey

    for f_info in files_info:
        if 'name' not in f_info:
            if not default_name:
                f_info['name'] = os.path.basename(f_info['path'])
            else:
                f_info['name'] = default_name

    # Add the genome fasta if asked
    if genome_fasta:
        if not create_dbkey:
            # delete the existing dbkey to force the recomputing of len file
            print("Deleting the dbkey entry before recreating it (needed to recompute the len file).")
            ctx.gi.tool_data.delete_data_table('__dbkeys__', "\t".join(dbkey_entry))

        if not genome_fasta_name:
            genome_fasta_name = default_name

        genome_fasta_abs = check_input([genome_fasta], check_existence=(not no_file_check), use_biomaj_env=(not no_biomaj_env))[0]

        # the dbkey is not (or not longer) existing: create it while adding the ref genome to force the computing of the len file
        print("Adding a new genome using fasta file '%s' -> '%s'" % (genome_fasta_name, genome_fasta_abs))
        params = {}
        params['dbkey_source|dbkey_source_selector'] = 'new'
        params['dbkey_source|dbkey'] = dbkey
        params['dbkey_source|dbkey_name'] = default_name
        params['sequence_name'] = genome_fasta_name
        params['reference_source|reference_source_selector'] = 'directory'
        params['reference_source|fasta_filename'] = genome_fasta_abs
        params['reference_source|create_symlink'] = 'true'
        params['sorting|sort_selector'] = fasta_sorting_method
        if fasta_sorting_method == 'custom':
            params['sorting|handle_not_listed|handle_not_listed_selector'] = fasta_custom_sort_handling
            n = 0
            for i in fasta_custom_sort_list.split(','):
                params['sorting|sequence_identifiers_' + n + '|identifier'] = i
                n += 1
        fetch_res = ctx.gi.tools.run_tool(None, ADD_FASTA_TOOL_ID, params)
        datasetid = fetch_res['outputs'][0]['id']
        jobid = None
        if 'jobs' in fetch_res:
            jobid = fetch_res['jobs'][0]['id']
        wait_completion(ctx.gi, datasetid, jobid)

    elif create_dbkey:  # Create the dbkey without ref genome (no len computing)
        print("Will create the dbkey '" + dbkey + "'")
        files_info.append({'table': '__dbkeys__', 'name': dbkey_display_name})
        table_counts['__dbkeys__'] = 1

    # Now add all associated data
    manual_dm_params = {}
    index_entry = 0
    for f_info in files_info:

        if 'path' in f_info:
            print("Adding a new entry to table '%s': '%s' -> '%s'" % (f_info['table'], f_info['name'], f_info['path']))
        else:
            print("Adding a new entry to table '%s': '%s' -> No path" % (f_info['table'], f_info['name']))

        vals = {
            'dbkey': dbkey,
            'name': f_info['name'],
            'path': f_info['path'] if 'path' in f_info else '',
            'db_path': f_info['path'] if 'path' in f_info else '',  # diamond data table
            'url': f_info['path'] if 'path' in f_info else '',
            'with-gtf': '1' if star_with_gtf else '0',  # rnastar data table, old data table
            'with_gene_model': '1' if star_with_gtf else '0',  # rnastar data table, recent data table
            'version': star_version if star_version else '0',  # rnastar data table, recent data table
            'len_path': f_info['path'] if 'path' in f_info else '',  # __dbkeys__data table
        }

        if dbkey and table_counts[f_info['table']] == 1:  # The id must be unique, only use dbkey if adding only one blastdb
            vals['value'] = dbkey
        else:
            vals['value'] = dbkey + "_" + str(uuid.uuid4())  # Let it be generated

        col_index = 0
        for col in tables_format[f_info['table']]:
            if col not in vals:
                warn("Skipping unknown column named '%s' in table '%s'." % (col, f_info['table']))

            manual_dm_params['data_tables_%s|columns_%s|data_table_column_name' % (index_entry, col_index)] = col
            manual_dm_params['data_tables_%s|columns_%s|data_table_column_value' % (index_entry, col_index)] = vals[col]
            manual_dm_params['data_tables_%s|columns_%s|is_path|is_path_selector' % (index_entry, col_index)] = 'no'
            col_index += 1

        manual_dm_params['data_tables_%s|data_table_name' % (index_entry)] = f_info['table']
        index_entry += 1

    fetch_res = ctx.gi.tools.run_tool(None, DM_MANUAL_TOOL_ID, manual_dm_params)
    datasetid = fetch_res['outputs'][0]['id']
    jobid = None
    if 'jobs' in fetch_res:
        jobid = fetch_res['jobs'][0]['id']
    wait_completion(ctx.gi, datasetid, jobid)

    # Reload all tables just in case
    time.sleep(1)  # Reloading too soon might not work for some strange reason
    for table in table_counts:
        print("Reloading table '%s'" % table)
        ctx.gi.tool_data.reload_data_table(table)



import time

from biomaj2galaxy import pass_context

import click


@click.command()
@click.argument("dbkey", type=str)
@click.argument("tables", nargs=-1)
@click.option(
    "--exact",
    help="Remove only exact matches instead of removing all lines with an id beginning with the given dbkey.",
    is_flag=True
)
@pass_context
def rm(ctx, dbkey, tables, exact):
    """Remove data from Galaxy data tables, where DBKEY is the id of the data to remove, and TABLES is an optional list of tables to remove data from (by default, data will be removed in all tables)."""

    # Define some simpler synonyms for data tables
    data_table_synonyms = {
        'fasta': 'all_fasta',
        'bowtie': 'bowtie_indexes',
        'bowtie2': 'bowtie2_indexes',
        'bwa': 'bwa_indexes',
        'bwa_mem': 'bwa_mem_indexes',
        'tophat2': 'tophat2_indexes',
    }

    # Fetch the list of known tables with their columns
    tables_format = {}
    tables_entries = {}
    online_tables = ctx.gi.tool_data.get_data_tables()
    for t in online_tables:
        content = ctx.gi.tool_data.show_data_table(t['name'])
        tables_format[t['name']] = content['columns']
        tables_entries[t['name']] = content['fields']

    tables_to_clean = []
    if tables:
        for table in tables:
            if table in data_table_synonyms:
                table = data_table_synonyms[table]

            if table in tables_format:
                tables_to_clean.append(table)
            else:
                raise Exception('Unknown data table name "%s"' % table)
    else:
        tables_to_clean = tables_format.keys()

    print("Will remove '%s' entries from tables: \n%s" % (dbkey, ', '.join(tables_to_clean)))

    # Always delete from the __dbkeys__ table
    table = '__dbkeys__'
    if 'value' in tables_format[table]:
        dbkey_field = tables_format[table].index('value')

        for line in tables_entries[table]:
            if line[dbkey_field] == dbkey:
                print("Deleting from '" + table + "' table")
                ctx.gi.tool_data.delete_data_table(table, "\t".join(line))

    # Remove from asked tables
    for table in tables_to_clean:
        if 'dbkey' in tables_format[table]:
            dbkey_field = tables_format[table].index('dbkey')
        elif 'value' in tables_format[table]:
            dbkey_field = tables_format[table].index('value')
        else:
            continue

        for line in tables_entries[table]:
            if len(line) > dbkey_field:  # Sometimes Galaxy is lying (featurecounts_anno table)
                if (exact and line[dbkey_field] == dbkey) or (not exact and line[dbkey_field].startswith(dbkey)):
                    print("Deleting from '" + table + "' table")
                    ctx.gi.tool_data.delete_data_table(table, "\t".join(line))

    # Reload all tables just in case
    time.sleep(1)  # Reloading too soon might not work for some strange reason
    print("Reloading tables")
    for table in tables_to_clean:
        ctx.gi.tool_data.reload_data_table(table)
    print("Done")
